@@ -29,6 +29,7 @@
 #include "wasm-printing.h"
 #include "wasm-s-parser.h"
 #include "wasm-validator.h"
+#include "wasm2asm.h"
 #include "cfg/Relooper.h"
 #include "ast_utils.h"
 #include "shell-interface.h"
@@ -474,6 +475,7 @@ BinaryenExpressionRef BinaryenCallIndirect(BinaryenModuleRef module, BinaryenExp
   }
   ret->fullType = type;
   ret->type = wasm->getFunctionType(ret->fullType)->result;
+  ret->finalize();
   return static_cast<Expression*>(ret);
 }
 BinaryenExpressionRef BinaryenGetLocal(BinaryenModuleRef module, BinaryenIndex index, BinaryenType type) {
@@ -651,7 +653,7 @@ BinaryenExpressionRef BinaryenDrop(BinaryenModuleRef module, BinaryenExpressionR
 }
 BinaryenExpressionRef BinaryenReturn(BinaryenModuleRef module, BinaryenExpressionRef value) {
   auto* ret = Builder(*((Module*)module)).makeReturn((Expression*)value);
- 
+
   if (tracing) {
     auto id = noteExpression(ret);
     std::cout << "  expressions[" << id << "] = BinaryenReturn(the_module, expressions[" << expressions[value] << "]);\n";
@@ -744,7 +746,7 @@ BinaryenFunctionRef BinaryenAddFunction(BinaryenModuleRef module, const char* na
   return ret;
 }
 
-BinaryenImportRef BinaryenAddGlobal(BinaryenModuleRef module, const char* name, BinaryenType type, bool mutable_, BinaryenExpressionRef init) {
+BinaryenImportRef BinaryenAddGlobal(BinaryenModuleRef module, const char* name, BinaryenType type, int8_t mutable_, BinaryenExpressionRef init) {
   if (tracing) {
     std::cout << "  BinaryenAddGlobal(the_module, \"" << name << "\", types[" << type << "], " << mutable_ << ", " << expressions[init] << ");\n";
   }
@@ -753,7 +755,7 @@ BinaryenImportRef BinaryenAddGlobal(BinaryenModuleRef module, const char* name,
   auto* ret = new Global();
   ret->name = name;
   ret->type = WasmType(type);
-  ret->mutable_ = mutable_;
+  ret->mutable_ = !!mutable_;
   ret->init = (Expression*)init;
   wasm->addGlobal(ret);
   return ret;
@@ -932,6 +934,21 @@ void BinaryenModulePrint(BinaryenModuleRef module) {
   WasmPrinter::printModule((Module*)module);
 }
 
+void BinaryenModulePrintAsmjs(BinaryenModuleRef module) {
+  if (tracing) {
+    std::cout << "  BinaryenModulePrintAsmjs(the_module);\n";
+  }
+
+  Module* wasm = (Module*)module;
+  Wasm2AsmBuilder::Flags builderFlags;
+  Wasm2AsmBuilder wasm2asm(builderFlags);
+  Ref asmjs = wasm2asm.processWasm(wasm);
+  JSPrinter jser(true, true, asmjs);
+  jser.printAst();
+
+  std::cout << jser.buffer;
+}
+
 int BinaryenModuleValidate(BinaryenModuleRef module) {
   if (tracing) {
     std::cout << "  BinaryenModuleValidate(the_module);\n";
