@@ -368,6 +368,7 @@ struct PrintSExpression : public Visitor<PrintSExpression> {
   }
   void visitAtomicRMW(AtomicRMW* curr) {
     o << '(';
+    prepareColor(o);
     printRMWSize(o, curr->type, curr->bytes);
     switch (curr->op) {
       case Add:  o << "add";  break;
@@ -388,6 +389,7 @@ struct PrintSExpression : public Visitor<PrintSExpression> {
   }
   void visitAtomicCmpxchg(AtomicCmpxchg* curr) {
     o << '(';
+    prepareColor(o);
     printRMWSize(o, curr->type, curr->bytes);
     o << "cmpxchg";
     restoreNormalColor(o);
@@ -400,6 +402,24 @@ struct PrintSExpression : public Visitor<PrintSExpression> {
     printFullLine(curr->replacement);
     decIndent();
   }
+  void visitAtomicWait(AtomicWait* curr) {
+    o << '(' ;
+    prepareColor(o);
+    o << printWasmType(curr->expectedType) << ".wait";
+    restoreNormalColor(o);
+    incIndent();
+    printFullLine(curr->ptr);
+    printFullLine(curr->expected);
+    printFullLine(curr->timeout);
+    decIndent();
+  }
+  void visitAtomicWake(AtomicWake* curr) {
+    printOpening(o, "wake");
+    incIndent();
+    printFullLine(curr->ptr);
+    printFullLine(curr->wakeCount);
+    decIndent();
+  }
   void visitConst(Const *curr) {
     o << curr->value;
   }
@@ -454,6 +474,11 @@ struct PrintSExpression : public Visitor<PrintSExpression> {
       case DemoteFloat64:          o << "f32.demote/f64"; break;
       case ReinterpretInt32:       o << "f32.reinterpret/i32"; break;
       case ReinterpretInt64:       o << "f64.reinterpret/i64"; break;
+      case ExtendS8Int32:          o << "i32.extend8_s"; break;
+      case ExtendS16Int32:         o << "i32.extend16_s"; break;
+      case ExtendS8Int64:          o << "i64.extend8_s"; break;
+      case ExtendS16Int64:         o << "i64.extend16_s"; break;
+      case ExtendS32Int64:         o << "i64.extend32_s"; break;
       default: abort();
     }
     incIndent();
@@ -562,8 +587,7 @@ struct PrintSExpression : public Visitor<PrintSExpression> {
     decIndent();
   }
   void visitDrop(Drop *curr) {
-    o << '(';
-    prepareColor(o) << "drop";
+    printOpening(o, "drop");
     incIndent();
     printFullLine(curr->value);
     decIndent();
