@@ -35,6 +35,11 @@
 
 namespace wasm {
 
+enum {
+  // the maximum amount of bytes we emit per LEB
+  MaxLEB32Bytes = 5
+};
+
 template<typename T, typename MiniT>
 struct LEB {
   static_assert(sizeof(MiniT) == 1, "MiniT must be a byte");
@@ -63,7 +68,9 @@ struct LEB {
     } while (more);
   }
 
-  void writeAt(std::vector<uint8_t>* out, size_t at, size_t minimum = 0) {
+  // @minimum: a minimum number of bytes to write, padding as necessary
+  // returns the number of bytes written
+  size_t writeAt(std::vector<uint8_t>* out, size_t at, size_t minimum = 0) {
     T temp = value;
     size_t offset = 0;
     bool more;
@@ -77,6 +84,7 @@ struct LEB {
       (*out)[at + offset] = byte;
       offset++;
     } while (more);
+    return offset;
   }
 
   void read(std::function<MiniT()> get) {
@@ -134,7 +142,7 @@ class BufferWithRandomAccess : public std::vector<uint8_t> {
   bool debug;
 
 public:
-  BufferWithRandomAccess(bool debug) : debug(debug) {}
+  BufferWithRandomAccess(bool debug = false) : debug(debug) {}
 
   BufferWithRandomAccess& operator<<(int8_t x) {
     if (debug) std::cerr << "writeInt8: " << (int)(uint8_t)x << " (at " << size() << ")" << std::endl;
@@ -258,15 +266,31 @@ public:
     (*this)[i+2] = x & 0xff; x >>= 8;
     (*this)[i+3] = x & 0xff;
   }
-  void writeAt(size_t i, U32LEB x) {
+
+  // writes out an LEB to an arbitrary location. this writes the LEB as a full
+  // 5 bytes, the fixed amount that can easily be set aside ahead of time
+  void writeAtFullFixedSize(size_t i, U32LEB x) {
     if (debug) std::cerr << "backpatchU32LEB: " << x.value << " (at " << i << ")" << std::endl;
-    x.writeAt(this, i, 5); // fill all 5 bytes, we have to do this when backpatching
+    x.writeAt(this, i, MaxLEB32Bytes); // fill all 5 bytes, we have to do this when backpatching
+  }
+  // writes out an LEB of normal size
+  // returns how many bytes were written
+  size_t writeAt(size_t i, U32LEB x) {
+    if (debug) std::cerr << "writeAtU32LEB: " << x.value << " (at " << i << ")" << std::endl;
+    return x.writeAt(this, i);
   }
 
   template <typename T>
   void writeTo(T& o) {
     for (auto c : *this) o << c;
   }
+
+  std::vector<char> getAsChars() {
+    std::vector<char> ret;
+    ret.resize(size());
+    std::copy(begin(), end(), ret.begin());
+    return ret;
+  }
 };
 
 namespace BinaryConsts {
@@ -507,10 +531,20 @@ enum ASTNodes {
   F32ReinterpretI32 = 0xbe,
   F64ReinterpretI64 = 0xbf,
 
+  I32ExtendS8 = 0xc0,
+  I32ExtendS16 = 0xc1,
+  I64ExtendS8 = 0xc2,
+  I64ExtendS16 = 0xc3,
+  I64ExtendS32 = 0xc4,
+
   AtomicPrefix = 0xfe
 };
 
 enum AtomicOpcodes {
+  AtomicWake = 0x00,
+  I32AtomicWait = 0x01,
+  I64AtomicWait = 0x02,
+
   I32AtomicLoad = 0x10,
   I64AtomicLoad = 0x11,
   I32AtomicLoad8U = 0x12,
@@ -640,7 +674,8 @@ public:
   void writeHeader();
   int32_t writeU32LEBPlaceholder();
   void writeResizableLimits(Address initial, Address maximum, bool hasMaximum, bool shared);
-  int32_t startSection(BinaryConsts::Section code);
+  template<typename T>
+  int32_t startSection(T code);
   void finishSection(int32_t start);
   int32_t startSubsection(BinaryConsts::UserSections::Subsection code);
   void finishSubsection(int32_t start);
@@ -733,6 +768,8 @@ public:
   void visitStore(Store *curr);
   void visitAtomicRMW(AtomicRMW *curr);
   void visitAtomicCmpxchg(AtomicCmpxchg *curr);
+  void visitAtomicWait(AtomicWait *curr);
+  void visitAtomicWake(AtomicWake *curr);
   void visitConst(Const *curr);
   void visitUnary(Unary *curr);
   void visitBinary(Binary *curr);
@@ -830,9 +867,15 @@ public:
 
   std::vector<Expression*> expressionStack;
 
+  bool definitelyUnreachable; // set when we know code is definitely unreachable. this helps parse
+                              // stacky wasm code, which can be unsuitable for our IR when unreachable
+
   BinaryConsts::ASTNodes lastSeparator = BinaryConsts::End;
 
+  // process a block-type scope, until an end or else marker, or the end of the function
   void processExpressions();
+  void skipUnreachableCode();
+
   Expression* popExpression();
   Expression* popNonVoidExpression();
 
@@ -861,9 +904,12 @@ public:
   int depth = 0; // only for debugging
 
   BinaryConsts::ASTNodes readExpression(Expression*& curr);
+  void pushBlockElements(Block* curr, size_t start, size_t end);
   void visitBlock(Block *curr);
-  Expression* getMaybeBlock(WasmType type);
-  Expression* getBlock(WasmType type);
+
+  // Gets a block of expressions. If it's just one, return that singleton.
+  Expression* getBlockOrSingleton(WasmType type);
+
   void visitIf(If *curr);
   void visitLoop(Loop *curr);
   BreakTarget getBreakTarget(int32_t offset);
@@ -892,6 +938,8 @@ public:
   bool maybeVisitStore(Expression*& out, uint8_t code, bool isAtomic);
   bool maybeVisitAtomicRMW(Expression*& out, uint8_t code);
   bool maybeVisitAtomicCmpxchg(Expression*& out, uint8_t code);
+  bool maybeVisitAtomicWait(Expression*& out, uint8_t code);
+  bool maybeVisitAtomicWake(Expression*& out, uint8_t code);
   bool maybeVisitConst(Expression*& out, uint8_t code);
   bool maybeVisitUnary(Expression*& out, uint8_t code);
   bool maybeVisitBinary(Expression*& out, uint8_t code);
