@@ -98,6 +98,10 @@ enum UnaryOp {
   PromoteFloat32, // f32 to f64
   DemoteFloat64, // f64 to f32
   ReinterpretInt32, ReinterpretInt64, // reinterpret bits to float
+  // The following sign-extention operators go along with wasm atomics support.
+  // Extend signed subword-sized integer. This differs from e.g. ExtendSInt32
+  // because the input integer is in an i64 value insetad of an i32 value.
+  ExtendS8Int32, ExtendS16Int32, ExtendS8Int64, ExtendS16Int64, ExtendS32Int64,
 };
 
 enum BinaryOp {
@@ -183,6 +187,8 @@ public:
     UnreachableId,
     AtomicCmpxchgId,
     AtomicRMWId,
+    AtomicWaitId,
+    AtomicWakeId,
     NumExpressionIds
   };
   Id _id;
@@ -456,6 +462,30 @@ class AtomicCmpxchg : public SpecificExpression<Expression::AtomicCmpxchgId> {
   void finalize();
 };
 
+class AtomicWait : public SpecificExpression<Expression::AtomicWaitId> {
+ public:
+  AtomicWait() = default;
+  AtomicWait(MixedArena& allocator) : AtomicWait() {}
+
+  Expression* ptr;
+  Expression* expected;
+  Expression* timeout;
+  WasmType expectedType;
+
+  void finalize();
+};
+
+class AtomicWake : public SpecificExpression<Expression::AtomicWakeId> {
+ public:
+  AtomicWake() = default;
+  AtomicWake(MixedArena& allocator) : AtomicWake() {}
+
+  Expression* ptr;
+  Expression* wakeCount;
+
+  void finalize();
+};
+
 class Const : public SpecificExpression<Expression::ConstId> {
 public:
   Const() {}
@@ -586,8 +616,8 @@ public:
   WasmType getLocalType(Index index);
 
   Name getLocalNameOrDefault(Index index);
+  Name getLocalNameOrGeneric(Index index);
 
-private:
   bool hasLocalName(Index index) const;
 };
 
@@ -744,6 +774,7 @@ public:
 
   void removeImport(Name name);
   void removeExport(Name name);
+  void removeFunction(Name name);
   // TODO: remove* for other elements
 
   void updateMaps();
