@@ -495,7 +495,7 @@ void SExpressionWasmBuilder::parseFunction(Element& s, bool preParseImport) {
         std::move(params),
         result,
         std::move(vars)
-                                                                        ));
+    ));
   };
   auto ensureAutoBlock = [&]() {
     if (!autoBlock) {
@@ -699,7 +699,12 @@ Expression* SExpressionWasmBuilder::makeExpression(Element& s) {
           if (op[2] == 0) return makeBinary(s, BINARY_INT_OR_FLOAT(Eq), type);
           if (op[2] == 'z') return makeUnary(s, type == i32 ? UnaryOp::EqZInt32 : UnaryOp::EqZInt64, type);
         }
-        if (op[1] == 'x') return makeUnary(s, op[7] == 'u' ? UnaryOp::ExtendUInt32 : UnaryOp::ExtendSInt32, type);
+        if (op[1] == 'x') {
+          if (op[6] == '8') return makeUnary(s, type == i32 ? UnaryOp::ExtendS8Int32 : UnaryOp::ExtendS8Int64, type);
+          if (op[6] == '1') return makeUnary(s, type == i32 ? UnaryOp::ExtendS16Int32 : UnaryOp::ExtendS16Int64, type);
+          if (op[6] == '3') return makeUnary(s, UnaryOp::ExtendS32Int64, type);
+          return makeUnary(s, op[7] == 'u' ? UnaryOp::ExtendUInt32 : UnaryOp::ExtendSInt32, type);
+        }
         abort_on(op);
       }
       case 'f': {
@@ -781,6 +786,7 @@ Expression* SExpressionWasmBuilder::makeExpression(Element& s) {
         abort_on(op);
       }
       case 'w': {
+        if (!strncmp(op, "wait", strlen("wait"))) return makeAtomicWait(s, type);
         if (op[1] == 'r') return makeUnary(s,  UnaryOp::WrapInt64, type);
         abort_on(op);
       }
@@ -866,6 +872,9 @@ Expression* SExpressionWasmBuilder::makeExpression(Element& s) {
         if (str[1] == 'n') return allocator.alloc<Unreachable>();
         abort_on(str);
       }
+      case 'w': {
+        if (!strncmp(str, "wake", strlen("wake"))) return makeAtomicWake(s);
+      }
       default: abort_on(str);
     }
   }
@@ -916,6 +925,8 @@ Expression* SExpressionWasmBuilder::makeUnary(Element& s, UnaryOp op, WasmType t
       break;
     }
     case ExtendSInt32: case ExtendUInt32:
+    case ExtendS8Int32: case ExtendS16Int32:
+    case ExtendS8Int64: case ExtendS16Int64: case ExtendS32Int64:
     case WrapInt64:
     case PromoteFloat32:
     case DemoteFloat64:
@@ -1241,6 +1252,26 @@ Expression* SExpressionWasmBuilder::makeAtomicCmpxchg(Element& s, WasmType type,
   return ret;
 }
 
+Expression* SExpressionWasmBuilder::makeAtomicWait(Element& s, WasmType type) {
+  auto ret = allocator.alloc<AtomicWait>();
+  ret->type = i32;
+  ret->expectedType = type;
+  ret->ptr = parseExpression(s[1]);
+  ret->expected = parseExpression(s[2]);
+  ret->timeout = parseExpression(s[3]);
+  ret->finalize();
+  return ret;
+}
+
+Expression* SExpressionWasmBuilder::makeAtomicWake(Element& s) {
+  auto ret = allocator.alloc<AtomicWake>();
+  ret->type = i32;
+  ret->ptr = parseExpression(s[1]);
+  ret->wakeCount = parseExpression(s[2]);
+  ret->finalize();
+  return ret;
+}
+
 Expression* SExpressionWasmBuilder::makeIf(Element& s) {
   auto ret = allocator.alloc<If>();
   Index i = 1;
