@@ -8,7 +8,11 @@ use std::process::Command;
 fn gen_bindings() {
     let bindings = bindgen::Builder::default()
         .header("wrapper.h")
+        // See https://github.com/rust-lang-nursery/rust-bindgen/issues/947
+        .trust_clang_mangling(false)
         .generate_comments(true)
+        // https://github.com/rust-lang-nursery/rust-bindgen/issues/947#issuecomment-327100002
+        .layout_tests(false)
         .generate()
         .expect("Unable to generate bindings");
 
@@ -27,6 +31,24 @@ fn main() {
 
     gen_bindings();
 
+    let target = env::var("TARGET").ok();
+    if target.map_or(false, |target| target.contains("emscripten")) {
+        let mut build_wasm_binaryen_args = vec![];
+        if get_debug() {
+            build_wasm_binaryen_args.push("-g");
+        }
+
+        let _ = Command::new("./build-binaryen-bc.sh")
+            .args(&build_wasm_binaryen_args)
+            .status()
+            .unwrap();
+
+        
+        println!("cargo:rustc-link-search=native={}", env::var("OUT_DIR").unwrap());
+        println!("cargo:rustc-link-lib=static=binaryen-c");
+        return;
+    }
+
     let dst = cmake::Config::new("binaryen")
         .define("BUILD_STATIC_LIB", "ON")
         .build();
@@ -63,3 +85,11 @@ fn get_cpp_stdlib() -> Option<String> {
         }
     })
 }
+
+// See https://github.com/alexcrichton/gcc-rs/blob/10871a0e40/src/lib.rs#L1501
+fn get_debug() -> bool {
+    match env::var("DEBUG").ok() {
+        Some(s) => s != "false",
+        None => false,
+    }
+}
